# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from .extension_matadata import ExtensionMetadata
from .types import ExtensionClassType


class RegisteredExtension(object):
  """A registered extension in extension manager."""

  def __init__(
      self,
      metadata: ExtensionMetadata,
      type: str,
      ext_class: ExtensionClassType,
  ):
    self.metadata = metadata
    self.type = type
    self.ext_class = ext_class
