# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

PACKAGE_NAME = 'ai-edge-model-explorer'
MODULE_NAME = 'model_explorer'
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 8080
DEFAULT_COLAB_HEIGHT = 850
DEFAULT_SETTINGS = {
    'const_element_count_limit': 16,
}
